from pbpstats.resources.boxscore.boxscore import Boxscore
from pbpstats.resources.enhanced_pbp.enhanced_pbp import EnhancedPbp
from pbpstats.resources.games.games import Games
from pbpstats.resources.pbp.pbp import Pbp
from pbpstats.resources.possessions.possessions import Possessions
from pbpstats.resources.shots.shots import Shots

__all__ = [
    "Boxscore",
    "EnhancedPbp",
    "Games",
    "Pbp",
    "Possessions",
    "Shots",
]



from pbpstats.resources.enhanced_pbp.stats_nba.enhanced_pbp_item import (
    StatsEnhancedPbpItem,
)
from pbpstats.resources.enhanced_pbp import Rebound, Turnover


class StatsRebound(Rebound, StatsEnhancedPbpItem):
    """
    Class for rebound events
    """

    def __init__(self, *args):
        super().__init__(*args)

    def get_offense_team_id(self):
        """
        returns team id for team on offense for the shot that was rebounded
        """
        if self.is_real_rebound:
            return self.missed_shot.team_id
        if (
            isinstance(self.previous_event, Turnover)
            and not self.previous_event.is_no_turnover
        ):
            # shot clock turnover has place holder rebound after turnover
            # this correct team starts next possession
            team_ids = list(self.current_players.keys())
            return (
                team_ids[0]
                if team_ids[1] == self.previous_event.team_id
                else team_ids[1]
            )
        return self.previous_event.get_offense_team_id()



from pbpstats.resources.enhanced_pbp.data_nba.enhanced_pbp_item import (
    DataEnhancedPbpItem,
)
from pbpstats.resources.enhanced_pbp import FreeThrow


class DataFreeThrow(FreeThrow, DataEnhancedPbpItem):
    """
    Class for free throw events
    """

    def __init__(self, *args):
        super().__init__(*args)

    @property
    def is_made(self):
        """
        returns True if shot was made, False otherwise
        """
        return " Missed" not in self.description



"""
The ``Possession`` class has some basic properties for handling possession data
"""
from itertools import groupby
from operator import itemgetter

import pbpstats
from pbpstats.resources.enhanced_pbp import (
    FieldGoal,
    FreeThrow,
    JumpBall,
    Rebound,
    Substitution,
    Timeout,
    Turnover,
)


class Possession(object):
    """
    Class for possession

    :param list events: list of
        :obj:`~pbpstats.resources.enhanced_pbp.enhanced_pbp_item.EnhancedPbpItem` items for possession,
        typically from a possession data loader
    """

    def __init__(self, events):
        self.game_id = events[0].game_id
        self.period = events[0].period
        self.events = events

    def __repr__(self):
        return (
            f"<{type(self).__name__} GameId: {self.game_id}, Period: {self.period}, "
            f"Number: {self.number}, StartTime: {self.start_time}, EndTime: {self.end_time}, "
            f"OffenseTeamId: {self.offense_team_id}>"
        )

    @property
    def data(self):
        """
        returns dict possession data
        """
        return self.__dict__

    @property
    def start_time(self):
        """
        returns the time remaining (MM:SS) in the period when the possession started
        """
        if not hasattr(self, "previous_possession") or self.previous_possession is None:
            return self.events[0].clock
        return self.previous_possession.events[-1].clock

    @property
    def end_time(self):
        """
        returns the time remaining (MM:SS) in the period when the possession ended
        """
        return self.events[-1].clock

    @property
    def start_score_margin(self):
        """
        returns the score margin from the perspective of the team on offense when the possession started
        """
        if not hasattr(self, "previous_possession") or self.previous_possession is None:
            score = self.events[0].score
        else:
            score = self.previous_possession.events[-1].score
        offense_team_id = self.offense_team_id
        offense_points = score[offense_team_id]
        defense_points = 0
        for team_id, points in score.items():
            if team_id != offense_team_id:
                defense_points = points
        return offense_points - defense_points

    def get_team_ids(self):
        """
        returns a list with the team ids of both teams playing
        """
        team_ids = list(
            set([event.team_id for event in self.events if event.team_id != 0])
        )
        prev_poss = self.previous_possession
        while len(team_ids) != 2 and prev_poss is not None:
            team_ids += [
                event.team_id for event in prev_poss.events if event.team_id != 0
            ]
            team_ids = list(set(team_ids))
            prev_poss = prev_poss.previous_possession
        next_poss = self.next_possession
        while len(team_ids) != 2 and next_poss is not None:
            team_ids += [
                event.team_id for event in next_poss.events if event.team_id != 0
            ]
            team_ids = list(set(team_ids))
            next_poss = next_poss.next_possession
        return team_ids

    @property
    def offense_team_id(self):
        """
        returns team id for team on offense on possession
        """
        if len(self.events) == 1 and isinstance(self.events[0], JumpBall):
            # if possession only has one event and it is a jump ball, need to check
            # how previous possession ended to see which team actually started with the ball
            # because team id on jump ball is team that won the jump ball
            prev_event = self.previous_possession_ending_event
            if isinstance(prev_event, Turnover) and not prev_event.is_no_turnover:
                team_ids = self.get_team_ids()
                return (
                    team_ids[0]
                    if team_ids[1] == prev_event.get_offense_team_id()
                    else team_ids[1]
                )
            if isinstance(prev_event, Rebound) and prev_event.is_real_rebound:
                if not prev_event.oreb:
                    team_ids = self.get_team_ids()
                    return (
                        team_ids[0]
                        if team_ids[1] == prev_event.get_offense_team_id()
                        else team_ids[1]
                    )
                return prev_event.get_offense_team_id()
            if isinstance(prev_event, (FieldGoal, FreeThrow)):
                if prev_event.is_made:
                    team_ids = self.get_team_ids()
                    return (
                        team_ids[0]
                        if team_ids[1] == prev_event.get_offense_team_id()
                        else team_ids[1]
                    )
                return prev_event.get_offense_team_id()
        return self.events[0].get_offense_team_id()

    @property
    def possession_has_timeout(self):
        """
        returns True if there was a timeout called on the current possession, False otherwise
        """
        for i, event in enumerate(self.events):
            if isinstance(event, Timeout) and event.clock != self.end_time:
                # timeout is not at possession end time
                if not (
                    event.next_event is not None
                    and (
                        isinstance(event.next_event, FreeThrow)
                        and not event.next_event.is_technical_ft
                    )
                    and event.clock == event.next_event.clock
                ):
                    # check to make sure timeout is not between/before FTs
                    return True
            elif isinstance(event, Timeout) and event.clock == self.end_time:
                timeout_time = event.clock
                after_timeout_index = i + 1
                # call time out and turn ball over at same time as timeout following time out
                for possession_event in self.events[after_timeout_index:]:
                    if (
                        isinstance(possession_event, Turnover)
                        and not possession_event.is_no_turnover
                        and possession_event.clock == timeout_time
                    ):
                        return True
        return False

    @property
    def previous_possession_has_timeout(self):
        """
        returns True if there was a timeout called at same time as possession ended, False otherwise
        """
        if self.previous_possession is not None:
            for event in self.previous_possession.events:
                if isinstance(event, Timeout) and event.clock == self.start_time:
                    if not (
                        event.next_event is not None
                        and isinstance(event.next_event, FreeThrow)
                        and event.clock == event.next_event.clock
                    ):
                        # check to make sure timeout is not beween FTs
                        return True
        return False

    @property
    def previous_possession_ending_event(self):
        """
        returns previous possession ending event - ignoring subs
        """
        previous_event_index = -1
        while isinstance(
            self.previous_possession.events[previous_event_index], Substitution
        ) and len(self.previous_possession.events) > abs(previous_event_index):
            previous_event_index -= 1
        return self.previous_possession.events[previous_event_index]

    @property
    def possession_start_type(self):
        """
        returns possession start type string
        """
        if self.number == 1:
            return pbpstats.OFF_DEADBALL_STRING
        if self.possession_has_timeout or self.previous_possession_has_timeout:
            return pbpstats.OFF_TIMEOUT_STRING
        previous_possession_ending_event = self.previous_possession_ending_event
        if (
            isinstance(previous_possession_ending_event, (FieldGoal, FreeThrow))
            and previous_possession_ending_event.is_made
        ):
            shot_type = previous_possession_ending_event.shot_type
            return f"Off{shot_type}{pbpstats.MAKE_STRING}"
        if isinstance(previous_possession_ending_event, Turnover):
            if previous_possession_ending_event.is_steal:
                return pbpstats.OFF_LIVE_BALL_TURNOVER_STRING
            return pbpstats.OFF_DEADBALL_STRING
        if isinstance(previous_possession_ending_event, Rebound):
            if previous_possession_ending_event.player1_id == 0:
                # team rebound
                return pbpstats.OFF_DEADBALL_STRING
            missed_shot = previous_possession_ending_event.missed_shot
            shot_type = missed_shot.shot_type
            if hasattr(missed_shot, "is_blocked") and missed_shot.is_blocked:
                return f"Off{shot_type}{pbpstats.BLOCK_STRING}"
            return f"Off{shot_type}{pbpstats.MISS_STRING}"

        if isinstance(previous_possession_ending_event, JumpBall):
            # jump balls tipped out of bounds have no player2_id and should be off deadball
            if not hasattr(previous_possession_ending_event, "player2_id"):
                return pbpstats.OFF_LIVE_BALL_TURNOVER_STRING
            else:
                return pbpstats.OFF_DEADBALL_STRING
        return pbpstats.OFF_DEADBALL_STRING

    @property
    def previous_possession_end_shooter_player_id(self):
        """
        returns player id of player who took shot (make or miss) that ended previous possession.
        returns 0 if previous possession did not end with made field goal or live ball rebound
        """
        if self.previous_possession is not None and not (
            self.possession_has_timeout or self.previous_possession_has_timeout
        ):
            previous_possession_ending_event = self.previous_possession_ending_event
            if (
                isinstance(previous_possession_ending_event, FieldGoal)
                and previous_possession_ending_event.is_made
            ):
                return previous_possession_ending_event.player1_id
            if isinstance(previous_possession_ending_event, Rebound):
                if previous_possession_ending_event.player1_id != 0:
                    missed_shot = previous_possession_ending_event.missed_shot
                    return missed_shot.player1_id
        return 0

    @property
    def previous_possession_end_rebound_player_id(self):
        """
        returns player id of player who got rebound that ended previous possession.
        returns 0 if previous possession did not end with a live ball rebound
        """
        if self.previous_possession is not None and not (
            self.possession_has_timeout or self.previous_possession_has_timeout
        ):
            previous_possession_ending_event = self.previous_possession_ending_event
            if isinstance(previous_possession_ending_event, Rebound):
                if previous_possession_ending_event.player1_id != 0:
                    return previous_possession_ending_event.player1_id
        return 0

    @property
    def previous_possession_end_turnover_player_id(self):
        """
        returns player id of player who turned ball over that ended previous possession.
        returns 0 if previous possession did not end with a live ball turnover
        """
        if self.previous_possession is not None and not (
            self.possession_has_timeout or self.previous_possession_has_timeout
        ):
            previous_possession_ending_event = self.previous_possession_ending_event
            if isinstance(previous_possession_ending_event, Turnover):
                if previous_possession_ending_event.is_steal:
                    return previous_possession_ending_event.player1_id
        return 0

    @property
    def previous_possession_end_steal_player_id(self):
        """
        returns player id of player who got steal that ended previous possession.
        returns 0 if previous possession did not end with a live ball turnover
        """
        if self.previous_possession is not None and not (
            self.possession_has_timeout or self.previous_possession_has_timeout
        ):
            previous_possession_ending_event = self.previous_possession_ending_event
            if isinstance(previous_possession_ending_event, Turnover):
                if previous_possession_ending_event.is_steal:
                    return previous_possession_ending_event.player3_id
        return 0

    @property
    def possession_stats(self):
        """
        returns list of dicts with aggregate stats for possession
        """
        grouper = itemgetter(
            "player_id",
            "team_id",
            "opponent_team_id",
            "lineup_id",
            "opponent_lineup_id",
            "stat_key",
        )
        results = []
        event_stats = [
            event_stat for event in self.events for event_stat in event.event_stats
        ]
        for key, group in groupby(sorted(event_stats, key=grouper), grouper):
            temp_dict = dict(
                zip(
                    [
                        "player_id",
                        "team_id",
                        "opponent_team_id",
                        "lineup_id",
                        "opponent_lineup_id",
                        "stat_key",
                    ],
                    key,
                )
            )
            temp_dict["stat_value"] = sum(item["stat_value"] for item in group)
            results.append(temp_dict)

        return results



from pbpstats.resources.enhanced_pbp.data_nba.enhanced_pbp_item import (
    DataEnhancedPbpItem,
)
from pbpstats.resources.enhanced_pbp import EndOfPeriod


class DataEndOfPeriod(EndOfPeriod, DataEnhancedPbpItem):
    """
    Class for end of period events
    """

    def __init__(self, *args):
        super().__init__(*args)



from pbpstats.resources.enhanced_pbp.stats_nba.enhanced_pbp_item import (
    StatsEnhancedPbpItem,
)
from pbpstats.resources.enhanced_pbp import FreeThrow


class StatsFreeThrow(FreeThrow, StatsEnhancedPbpItem):
    """
    Class for free throw events
    """

    def __init__(self, *args):
        super().__init__(*args)

    @property
    def is_made(self):
        """
        returns True if shot was made, False otherwise
        """
        return "MISS " not in self.description

    def get_offense_team_id(self):
        """
        returns team id that took the shot
        """
        return self.team_id



OVERRIDES = {"pid": "player_id"}


class DataNbaBoxscoreItem(object):
    """
    Class for boxscore items from data.nba.com

    :param dict item: dict with boxscore stats from response
    :param int team_id: (optional) team id is not in dict with stats and can be added in here
    :param str team_abbreviation: (optional) team abbreviation is not in dict with stats and can be added in here
    """

    def __init__(self, item, team_id=None, team_abbreviation=None):
        self.team_id = team_id
        self.team_abbreviation = team_abbreviation
        for key, value in item.items():
            setattr(self, OVERRIDES.get(key, key), value)
        if hasattr(self, "fn") and hasattr(self, "ln"):
            self.name = f"{self.fn} {self.ln}"

    @property
    def data(self):
        """
        returns boxscore data dict
        """
        return self.__dict__



from pbpstats.resources.enhanced_pbp.data_nba.enhanced_pbp_item import (
    DataEnhancedPbpItem,
)
from pbpstats.resources.enhanced_pbp import Foul


class DataFoul(Foul, DataEnhancedPbpItem):
    """
    Class for foul events
    """

    def __init__(self, *args):
        super().__init__(*args)

    @property
    def number_of_fta_for_foul(self):
        """
        returns the number of free throws resulting from the foul
        """
        if "(1 FTA)" in self.description:
            return 1
        elif "(2 FTA)" in self.description:
            return 2
        elif "(3 FTA)" in self.description:
            return 3



"""
The ``Boxscore`` class has some basic properties for handling boxscore data
"""
from pbpstats.resources.base import Base


class Boxscore(Base):
    """
    Class for boxscore items

    :param list items: list of either
        :obj:`~pbpstats.resources.boxscore.stats_nba_boxscore_item.StatsNbaBoxscoreItem` or
        :obj:`~pbpstats.resources.boxscore.data_nba_boxscore_item.DataNbaBoxscoreItem` items,
        typically from a boxscore data loader
    """

    def __init__(self, items):
        self.items = items

    @property
    def data(self):
        """
        returns dict with boxscore items split up by player and team
        """
        return {"player": self.player_items, "team": self.team_items}

    @property
    def player_items(self):
        """
        returns list of player boxscore items
        """
        return [item.data for item in self.items if hasattr(item, "player_id")]

    @property
    def team_items(self):
        """
        returns list of team boxscore items
        """
        return [item.data for item in self.items if not hasattr(item, "player_id")]

    @property
    def player_name_map(self):
        """
        returns dict mapping player id to player name
        """
        return {item["player_id"]: item["name"] for item in self.player_items}

    @property
    def player_team_map(self):
        """
        returns dict mapping player id to team id
        """
        return {item["player_id"]: item["team_id"] for item in self.player_items}



"""
``StatsNbaPossessionLoader`` loads possession data for a game and
creates :obj:`~pbpstats.resources.possessions.possession.Possession` objects for each possession

The following code will load possession data for game id "0021900001" from a
pbp file located in the ``/pbp`` subdirectory of the ``/data`` directory

.. code-block:: python

    from pbpstats.data_loader import StatsNbaPossessionLoader

    possession_loader = StatsNbaPossessionLoader("0021900001", "file", "/data")
    print(possession_loader.items[0].data)  # prints dict with the first possession of the game
"""
import os
import json

from pbpstats import (
    NBA_GAME_ID_PREFIX,
    G_LEAGUE_GAME_ID_PREFIX,
    WNBA_GAME_ID_PREFIX,
    NBA_STRING,
    G_LEAGUE_STRING,
    WNBA_STRING,
)
from pbpstats.overrides import IntDecoder
from pbpstats.data_loader.stats_nba.enhanced_pbp_loader import StatsNbaEnhancedPbpLoader
from pbpstats.data_loader.nba_possession_loader import NbaPossessionLoader
from pbpstats.resources.possessions.possession import Possession
from pbpstats.resources.enhanced_pbp import Foul


class TeamHasBackToBackPossessionsException(Exception):
    """
    Class for exception when a team is credited with back-to-back possessions.

    You can manually edit the event order in the pbp file stored on disk or add
    an event to the overrides file in your data directory to fix this.
    """

    pass


class StatsNbaPossessionLoader(NbaPossessionLoader):
    """
    Loads stats.nba.com source possession data for game.
    Possessions are stored in items attribute as :obj:`~pbpstats.resources.possessions.possession.Possession` objects

    :param str game_id: NBA Stats Game Id
    :param str source: Where should data be loaded from. Options are 'web' or 'file'
    :param str file_directory: (optional if source is 'web')
        Directory in which data should be either stored (if source is web) or loaded from (if source is file).
        The specific file location will be `stats_<game_id>.json` in the `/pbp` subdirectory.
        If not provided response data will not be saved on disk.
    :raises: :obj:`~pbpstats.data_loader.stats_nba.possessions_loader.TeamHasBackToBackPossessionsException`:
        If team has the ball on back-to-back possessions.
    """

    data_provider = "stats_nba"
    resource = "Possessions"
    parent_object = "Game"

    def __init__(self, game_id, source, file_directory=None):

        self.file_directory = file_directory
        self.game_id = game_id
        pbp_events = StatsNbaEnhancedPbpLoader(game_id, source, file_directory)
        self.events = pbp_events.items
        events_by_possession = self._split_events_by_possession()
        self.items = [
            Possession(possession_events) for possession_events in events_by_possession
        ]
        self._add_extra_attrs_to_all_possessions()
        self._load_bad_possession_overrides()
        self._check_that_possessions_alternate()

    @property
    def league(self):
        """
        Returns League for game id.

        First 2 in game id represent league - 00 for nba, 10 for wnba, 20 for g-league
        """
        if self.game_id[0:2] == NBA_GAME_ID_PREFIX:
            return NBA_STRING
        elif self.game_id[0:2] == G_LEAGUE_GAME_ID_PREFIX:
            return G_LEAGUE_STRING
        elif self.game_id[0:2] == WNBA_GAME_ID_PREFIX:
            return WNBA_STRING

    def _check_that_possessions_alternate(self):
        """
        checks that a team doesn't have back-to-back possessions
        usually caused by pbp events being out of order that can be fixed manually
        """
        for possession in self.items:
            if possession.previous_possession is not None:
                poss = possession
                prev_poss = possession.previous_possession
            elif possession.next_possession is not None:
                poss = possession.next_possession
                prev_poss = possession
            if poss.offense_team_id == prev_poss.offense_team_id:
                game_id = (
                    self.game_id if self.league == NBA_STRING else int(self.game_id)
                )
                if not (
                    game_id in self.bad_pbp_cases.keys()
                    and poss.period in self.bad_pbp_cases[game_id].keys()
                    and poss.number in self.bad_pbp_cases[game_id][poss.period]
                ):
                    ignore_because_of_flagrant = False
                    events_to_check = [event for event in prev_poss.events]
                    if prev_poss.previous_possession is not None:
                        events_to_check += prev_poss.previous_possession.events
                    for event in events_to_check:
                        if isinstance(event, Foul) and event.is_flagrant:
                            ignore_because_of_flagrant = True

                    if not ignore_because_of_flagrant:
                        exception_text = (
                            f"GameId: {poss.game_id}, Period: {poss.period}, "
                            f"Number: {poss.number}, Events: {poss.events}, "
                            f"Previous Events: {prev_poss.events}>"
                        )

                        raise TeamHasBackToBackPossessionsException(exception_text)

    def _load_bad_possession_overrides(self):
        self.bad_pbp_cases = {}
        if self.file_directory is not None:
            bad_pbp_possessions_file_path = (
                f"{self.file_directory}/overrides/bad_pbp_possessions.json"
            )
            if os.path.isfile(bad_pbp_possessions_file_path):
                with open(bad_pbp_possessions_file_path) as f:
                    # bad pbp where event is missing in pbp causing back to back possessions for same team - this will prevent back to back possession exception from being raised
                    # {GameId: {Period:[EventNum]}}
                    self.bad_pbp_cases = json.loads(f.read(), cls=IntDecoder)



"""
``StatsNbaScoreboardLoader`` loads all games for a date and
creates :obj:`~pbpstats.resources.games.stats_nba_game_item.StatsNbaGameItem`
objects for each game

The following code will load data for 02/03/2020

.. code-block:: python

    from pbpstats.data_loader import StatsNbaScoreboardLoader

    game_finder_loader = StatsNbaScoreboardLoader("02/03/2020", "nba", "/data, "file")
    print(game_finder_loader.items[0].data) # prints dict for first game
"""
import json
import os

from pbpstats import (
    NBA_STRING,
    G_LEAGUE_STRING,
    WNBA_STRING,
    NBA_GAME_ID_PREFIX,
    G_LEAGUE_GAME_ID_PREFIX,
    WNBA_GAME_ID_PREFIX,
)
from pbpstats.data_loader.abs_data_loader import check_file_directory
from pbpstats.data_loader.stats_nba.file_loader import StatsNbaFileLoader
from pbpstats.data_loader.stats_nba.web_loader import StatsNbaWebLoader
from pbpstats.resources.games.stats_nba_game_item import StatsNbaGameItem


class StatsNbaScoreboardLoader(StatsNbaFileLoader, StatsNbaWebLoader):
    """
    Loads stats.nba.com source data for date.
    Games are stored in items attribute
    as :obj:`~pbpstats.resources.games.stats_nba_game_item.StatsNbaGameItem` objects

    :param str date: Formatted as MM/DD/YYYY
    :param str league_string: Options are 'nba', 'wnba' or 'gleague'
    :param str file_directory: (optional if source is 'web')
        Directory in which data should be either stored (if source is web) or loaded from (if source is file).
        The specific file location will be `stats_<league>_<date>.json` in the `/schedule` subdirectory.
        If None response data will not be saved on disk.
    :param str source: Where should data be loaded from. Options are 'web' or 'file'
    """

    data_provider = "stats_nba"
    resource = "Games"
    parent_object = "Day"

    def __init__(self, date, league_string, file_directory, source):
        self.date = date
        self.league_string = league_string
        self.file_directory = file_directory
        self.source = source
        self._load_data()
        self._make_scoreboard_items()

    def _load_data(self):
        source_method = getattr(self, f"_from_{self.source}")
        source_method()

    @check_file_directory
    def _from_file(self):
        self.file_path = f'{self.file_directory}/schedule/stats_{self.league_string}_{self.date.replace("/", "_")}.json'
        self._load_data_from_file()

    def _from_web(self):
        league_url_part = (
            f"{G_LEAGUE_STRING}.{NBA_STRING}"
            if self.league_string == G_LEAGUE_STRING
            else self.league_string
        )
        self.base_url = f"https://stats.{league_url_part}.com/stats/scoreboardV2"
        self.parameters = {
            "DayOffset": 0,
            "LeagueID": self.league_id,
            "gameDate": self.date,
        }
        self._load_request_data()

    def _save_data_to_file(self):
        if self.file_directory is not None and os.path.isdir(self.file_directory):
            file_path = f'{self.file_directory}/schedule/stats_{self.league_string}_{self.date.replace("/", "_")}.json'
            with open(file_path, "w") as outfile:
                json.dump(self.source_data, outfile)

    @property
    def league_id(self):
        """
        Returns League Id for league.

        00 for nba, 10 for wnba, 20 for g-league
        """
        if self.league_string == NBA_STRING:
            return NBA_GAME_ID_PREFIX
        elif self.league_string == WNBA_STRING:
            return WNBA_GAME_ID_PREFIX
        elif self.league_string == G_LEAGUE_STRING:
            return G_LEAGUE_GAME_ID_PREFIX

    def _make_scoreboard_items(self):
        self.items = [StatsNbaGameItem(item) for item in self.data]



import pbpstats


class Foul(object):
    """
    Class for foul events
    """

    event_type = 6

    @property
    def is_personal_foul(self):
        return self.event_action_type == 1

    @property
    def is_shooting_foul(self):
        return self.event_action_type == 2

    @property
    def is_loose_ball_foul(self):
        return self.event_action_type == 3

    @property
    def is_offensive_foul(self):
        return self.event_action_type == 4

    @property
    def is_inbound_foul(self):
        return self.event_action_type == 5

    @property
    def is_away_from_play_foul(self):
        return self.event_action_type == 6

    @property
    def is_clear_path_foul(self):
        return self.event_action_type == 9

    @property
    def is_double_foul(self):
        return self.event_action_type == 10

    @property
    def is_technical(self):
        return self.event_action_type in [11, 12, 13, 18, 19, 25, 30]

    @property
    def is_flagrant(self):
        return self.event_action_type in [14, 15]

    @property
    def is_double_technical(self):
        return self.event_action_type == 16

    @property
    def is_defensive_3_seconds(self):
        return self.event_action_type == 17

    @property
    def is_delay_of_game(self):
        return self.event_action_type == 18

    @property
    def is_charge(self):
        return self.event_action_type == 26

    @property
    def is_personal_block_foul(self):
        return self.event_action_type == 27

    @property
    def is_personal_take_foul(self):
        return self.event_action_type == 28

    @property
    def is_shooting_block_foul(self):
        return self.event_action_type == 29

    @property
    def counts_towards_penalty(self):
        """
        returns True if foul is a foul type that counts towards the penalty, False otherwise
        """
        return self.event_action_type in [1, 2, 3, 5, 6, 9, 14, 15, 27, 28, 29]

    @property
    def counts_as_personal_foul(self):
        """
        returns True if fouls is a foul type that counts as a personal foul, False otherwise
        """
        return self.event_action_type in [
            1,
            2,
            3,
            4,
            5,
            6,
            9,
            10,
            14,
            15,
            26,
            27,
            28,
            29,
        ]

    @property
    def foul_type_string(self):
        """
        returns string description of foul type
        """
        if self.is_personal_foul:
            return pbpstats.PERSONAL_FOUL_TYPE_STRING
        if self.is_shooting_foul:
            return pbpstats.SHOOTING_FOUL_TYPE_STRING
        if self.is_loose_ball_foul:
            return pbpstats.LOOSE_BALL_FOUL_TYPE_STRING
        if self.is_offensive_foul:
            return pbpstats.OFFENSIVE_FOUL_TYPE_STRING
        if self.is_inbound_foul:
            return pbpstats.INBOUND_FOUL_TYPE_STRING
        if self.is_away_from_play_foul:
            return pbpstats.AWAY_FROM_PLAY_FOUL_TYPE_STRING
        if self.is_clear_path_foul:
            return pbpstats.CLEAR_PATH_FOUL_TYPE_STRING
        if self.is_double_foul:
            return pbpstats.DOUBLE_FOUL_TYPE_STRING
        if self.event_action_type == 14:
            return pbpstats.FLAGRANT_1_FOUL_TYPE_STRING
        if self.event_action_type == 15:
            return pbpstats.FLAGRANT_2_FOUL_TYPE_STRING
        if self.is_defensive_3_seconds:
            return pbpstats.DEFENSIVE_3_SECONDS_FOUL_TYPE_STRING
        if self.is_charge:
            return pbpstats.CHARGE_FOUL_TYPE_STRING
        if self.is_personal_block_foul:
            return pbpstats.PERSONAL_BLOCK_TYPE_STRING
        if self.is_personal_take_foul:
            return pbpstats.PERSONAL_TAKE_TYPE_STRING
        if self.is_shooting_block_foul:
            return pbpstats.SHOOTING_BLOCK_TYPE_STRING

    @property
    def event_stats(self):
        """
        returns list of dicts with all stats for event
        """
        stats = []
        foul_type = self.foul_type_string
        is_penalty_event = self.is_penalty_event()
        if foul_type is not None:
            stats.append(
                {
                    "player_id": self.player1_id,
                    "team_id": self.team_id,
                    "stat_key": foul_type,
                    "stat_value": 1,
                }
            )
            if is_penalty_event:
                stats.append(
                    {
                        "player_id": self.player1_id,
                        "team_id": self.team_id,
                        "stat_key": f"{pbpstats.PENALTY_STRING}{foul_type}",
                        "stat_value": 1,
                    }
                )
            team_ids = list(self.current_players.keys())
            if hasattr(self, "player3_id"):
                p3_stat_key = (
                    foul_type
                    if self.is_double_foul
                    else foul_type + pbpstats.FOULS_DRAWN_TYPE_STRING
                )
                opponent_team_id = (
                    team_ids[0] if self.team_id == team_ids[1] else team_ids[1]
                )
                if self.player3_id in self.current_players[self.team_id]:
                    stats.append(
                        {
                            "player_id": self.player3_id,
                            "team_id": self.team_id,
                            "stat_key": p3_stat_key,
                            "stat_value": 1,
                        }
                    )
                    if is_penalty_event:
                        stats.append(
                            {
                                "player_id": self.player3_id,
                                "team_id": self.team_id,
                                "stat_key": f"{pbpstats.PENALTY_STRING}{p3_stat_key}",
                                "stat_value": 1,
                            }
                        )
                elif self.player3_id in self.current_players[opponent_team_id]:
                    stats.append(
                        {
                            "player_id": self.player3_id,
                            "team_id": opponent_team_id,
                            "stat_key": p3_stat_key,
                            "stat_value": 1,
                        }
                    )
                    if is_penalty_event:
                        stats.append(
                            {
                                "player_id": self.player3_id,
                                "team_id": opponent_team_id,
                                "stat_key": f"{pbpstats.PENALTY_STRING}{p3_stat_key}",
                                "stat_value": 1,
                            }
                        )

            lineups_ids = self.lineup_ids
            for stat in stats:
                opponent_team_id = (
                    team_ids[0] if stat["team_id"] == team_ids[1] else team_ids[1]
                )
                stat["lineup_id"] = lineups_ids[stat["team_id"]]
                stat["opponent_team_id"] = opponent_team_id
                stat["opponent_lineup_id"] = lineups_ids[opponent_team_id]
        return self.base_stats + stats



"""
``DataNbaScheduleLoader`` loads schedule data for a season and
creates :obj:`~pbpstats.resources.games.data_nba_game_item.DataNbaGameItem` objects for each game

The following code will load schedule data for 2019-20 NBA Regular Season

.. code-block:: python

    from pbpstats.data_loader import DataNbaScheduleLoader

    schedule_loader = DataNbaScheduleLoader("nba", "2019-20", "Regular Season", "web")
    print(schedule_loader.items[0].data)  # prints dict with the first game of the season
"""
import json
import os

from pbpstats import (
    NBA_STRING,
    G_LEAGUE_STRING,
    D_LEAGUE_STRING,
    WNBA_STRING,
    PLAYOFFS_STRING,
    REGULAR_SEASON_STRING,
    PLAY_IN_STRING,
    NBA_GAME_ID_PREFIX,
    G_LEAGUE_GAME_ID_PREFIX,
    WNBA_GAME_ID_PREFIX,
)
from pbpstats.data_loader.abs_data_loader import check_file_directory
from pbpstats.data_loader.data_nba.file_loader import DataNbaFileLoader
from pbpstats.data_loader.data_nba.web_loader import DataNbaWebLoader
from pbpstats.resources.games.data_nba_game_item import DataNbaGameItem


class DataNbaScheduleLoader(DataNbaFileLoader, DataNbaWebLoader):
    """
    Loads data.nba.com source schedule data for season.
    Games are stored in items attribute
    as :obj:`~pbpstats.resources.games.data_nba_game_item.DataNbaGameItem` objects

    :param str league: Options are 'nba', 'wnba' or 'gleague'
    :param str season: Can be formatted as either 2019-20 or 2019.
    :param str season_type: Options are 'Regular Season' or 'Playoffs' or 'Play In'
    :param str source: Where should data be loaded from. Options are 'web' or 'file'
    :param str file_directory: (optional if source is 'web')
        Directory in which data should be either stored (if source is web) or loaded from (if source is file).
        The specific file location will be `data_<league>_<season_year>.json` in the `/schedule` subdirectory.
        If not provided response data will not be saved on disk.
    """

    data_provider = "data_nba"
    resource = "Games"
    parent_object = "Season"

    def __init__(self, league, season, season_type, source, file_directory=None):
        self.league_string = league
        self.season_year = season.split("-")[0]
        self.file_directory = file_directory
        self.source = source
        self._load_data()
        self._make_game_items(season_type)

    def _load_data(self):
        source_method = getattr(self, f"_from_{self.source}")
        source_method()

    @check_file_directory
    def _from_file(self):
        self.file_path = f"{self.file_directory}/schedule/data_{self.league_string}_{self.season_year}.json"
        self._load_data_from_file()

    def _from_web(self):
        league_url_part = (
            NBA_STRING if self.league_string == G_LEAGUE_STRING else self.league_string
        )
        league_part = (
            D_LEAGUE_STRING
            if self.league_string == G_LEAGUE_STRING
            else self.league_string
        )
        self.url = f"https://data.{league_url_part}.com/data/10s/v2015/json/mobile_teams/{league_part}/{self.season_year}/league/{self.league_id}_full_schedule.json"
        self._load_request_data()

    def _save_data_to_file(self):
        if self.file_directory is not None and os.path.isdir(self.file_directory):
            file_path = f"{self.file_directory}/schedule/data_{self.league_string}_{self.season_year}.json"
            with open(file_path, "w") as outfile:
                json.dump(self.source_data, outfile)

    def _make_game_items(self, season_type):
        self.season_type = season_type
        self.items = [
            DataNbaGameItem(game)
            for item in self.data
            for game in item["mscd"]["g"]
            if self._is_season_type(game)
        ]

    def _is_season_type(self, game):
        if game["gid"][2] == "4" and self.season_type == PLAYOFFS_STRING:
            return True
        elif game["gid"][2] == "2" and self.season_type == REGULAR_SEASON_STRING:
            return True
        elif game["gid"][2] == "5" and self.season_type == PLAY_IN_STRING:
            return True
        return False

    @property
    def data(self):
        """
        returns raw JSON response data
        """
        return self.source_data["lscd"]

    @property
    def league_id(self):
        """
        Returns League Id for league.

        00 for nba, 10 for wnba, 20 for g-league
        """
        if self.league_string == NBA_STRING:
            return NBA_GAME_ID_PREFIX
        elif self.league_string == WNBA_STRING:
            return WNBA_GAME_ID_PREFIX
        elif self.league_string == G_LEAGUE_STRING:
            return G_LEAGUE_GAME_ID_PREFIX



from pbpstats.resources.enhanced_pbp.stats_nba.enhanced_pbp_item import (
    StatsEnhancedPbpItem,
)
from pbpstats.resources.enhanced_pbp import FieldGoal


class StatsFieldGoal(FieldGoal, StatsEnhancedPbpItem):
    """
    Class for field goal events
    """

    def __init__(self, *args):
        super().__init__(*args)

    @property
    def shot_value(self):
        """
        returns 3 if shot is a 3 point attempt, 2 otherwise
        """
        return 3 if " 3PT " in self.description else 2

    def get_offense_team_id(self):
        """
        returns team id that took the shot
        """
        return self.team_id



VIDEO_EVENT_ASSET_BASE_URL = "https://stats.nba.com/stats/videoeventsasset"
TODAYS_NBA_SCORES_URL = "http://data.nba.com/data/5s/v2015/json/mobile_teams/nba/2019/scores/00_todays_scores.json"
TODAYS_WNBA_SCORES_URL = "http://data.wnba.com/data/5s/v2015/json/mobile_teams/wnba/2020/scores/10_todays_scores.json"
TODAYS_G_LEAGUE_SCORES_URL = "https://data.nba.com/data/10s/v2015/json/mobile_teams/dleague/2019/scores/20_todays_scores.json"
REQUEST_TIMEOUT = 10
USER_AGENT = "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_9_3) AppleWebKit/537.75.14 (KHTML, like Gecko) Version/7.0.3 Safari/7046A194A"
REFERER = "http://stats.nba.com/"
HEADERS = {
    "User-Agent": USER_AGENT,
    "Referer": REFERER,
    "Accept": "application/json, text/plain, */*",
    "Accept-Language": "en-US,en;q=0.5",
    "x-nba-stats-origin": "stats",
    "x-nba-stats-token": "true",
    "DNT": "1",
    "Connection": "keep-alive",
    "Pragma": "no-cache",
    "Cache-Control": "no-cache",
}

NBA_STRING = "nba"
G_LEAGUE_STRING = "gleague"
WNBA_STRING = "wnba"
D_LEAGUE_STRING = "dleague"

NBA_GAME_ID_PREFIX = "00"
G_LEAGUE_GAME_ID_PREFIX = "20"
WNBA_GAME_ID_PREFIX = "10"

REGULAR_SEASON_STRING = "Regular Season"
PLAYOFFS_STRING = "Playoffs"
PLAY_IN_STRING = "Play In"

OFFENSIVE_ABBREVIATION_PREFIX = "Off"
DEFENSIVE_ABBREVIATION_PREFIX = "Def"
SECONDS_PLAYED_STRING = "SecondsPlayed"
SECONDS_PLAYED_OFFENSE_STRING = (
    f"{SECONDS_PLAYED_STRING}{OFFENSIVE_ABBREVIATION_PREFIX}"
)
SECONDS_PLAYED_DEFENSE_STRING = (
    f"{SECONDS_PLAYED_STRING}{DEFENSIVE_ABBREVIATION_PREFIX}"
)
OFFENSIVE_POSSESSION_STRING = "OffPoss"
DEFENSIVE_POSSESSION_STRING = "DefPoss"
SECOND_CHANCE_STRING = "SecondChance"
SECOND_CHANCE_SECONDS_PLAYED_OFFENSE_STRING = (
    f"{SECOND_CHANCE_STRING}{SECONDS_PLAYED_OFFENSE_STRING}"
)
SECOND_CHANCE_SECONDS_PLAYED_DEFENSE_STRING = (
    f"{SECOND_CHANCE_STRING}{SECONDS_PLAYED_DEFENSE_STRING}"
)

AT_RIM_CUTOFF = 4
SHORT_MID_RANGE_CUTOFF = 14
HEAVE_DISTANCE_CUTOFF = 40
HEAVE_TIME_CUTOFF = 2

TEAM_STAT_PLAYER_ID = (
    "0"  # used as player id for team stats (ex team rebounds, shot clock violations)
)

AT_RIM_STRING = "AtRim"
SHORT_MID_RANGE_STRING = "ShortMidRange"
LONG_MID_RANGE_STRING = "LongMidRange"
CORNER_3_STRING = "Corner3"
ARC_3_STRING = "Arc3"
UNKNOWN_SHOT_DISTANCE_STRING = "UnknownDistance2pt"
FREE_THROW_STRING = "FT"
FTS_MADE_STRING = "FtsMade"
FT_1_PT_MADE_STRING = "1ptFtsMade"
FT_2_PT_MADE_STRING = "2ptFtsMade"
FT_3_PT_MADE_STRING = "3ptFtsMade"
TECHNICAL_FTS_MADE_STRING = "TechFtsMade"
FTS_MISSED_STRING = "FtsMissed"
FT_1_PT_MISSED_STRING = "1ptFtsMissed"
FT_2_PT_MISSED_STRING = "2ptFtsMissed"
FT_3_PT_MISSED_STRING = "3ptFtsMissed"
ASSISTS_STRING = "Assists"
REBOUNDS_STRING = "Rebounds"
ASSISTED_STRING = "Assisted"
UNASSISTED_STRING = "Unassisted"
MISSED_STRING = "Missed"
BLOCKED_STRING = "Blocked"
REBOUND_OPPORTUNITIES_STRING = "ReboundOpportunities"
REBOUNDED_STRING = "Rebounded"
REBOUNDED_OPPORTUNITIES_STRING = (
    "ReboundedOpportunities"  # this is for getting oreb% on a player's missed shots
)
ON_FLOOR_OFFENSIVE_REBOUND_STRING = "OnFloorOffReb"  # used for calculating player usage
BLOCK_STRING = "Block"
SELF_REBOUND_STRING = "SelfOReb"

PUTBACKS_STRING = "Putbacks"
HEAVE_MISSES_STRING = "HeaveMisses"
HEAVE_MAKES_STRING = "HeaveMakes"

# attributes for shot data
SECONDS_SINCE_OREB_STRING = "SecondsSinceOReb"
OREB_SHOT_PLAYER_ID_STRING = "OrebShotPlayerId"
OREB_REBOUND_PLAYER_ID_STRING = "OrebReboundPlayerId"
START_SCORE_DIFFERENTIAL_STRING = "StartScoreDifferential"
SHOT_DISTANCE_STRING = "ShotDistance"
SECONDS_SINCE_PLAY_STARTED = "SecondsSincePlayStarted"
IS_REGULAR_SEASON_STRING = "IsRegularSeason"
SHOT_ANGLE_STRING = "ShotAngle"
PUTBACK_STRING = "Putback"

MAKE_STRING = "Make"
MISS_STRING = "Miss"
MADE_STRING = "Made"

SHOT_VALUE_STRING = "ShotValue"
OPPONENT_POINTS = "OpponentPoints"

PERSONAL_FOUL_TYPE_STRING = "Personal Fouls"
SHOOTING_FOUL_TYPE_STRING = "Shooting Fouls"
LOOSE_BALL_FOUL_TYPE_STRING = "Loose Ball Fouls"
OFFENSIVE_FOUL_TYPE_STRING = "Offensive Fouls"
AWAY_FROM_PLAY_FOUL_TYPE_STRING = "Away From Play Fouls"
CHARGE_FOUL_TYPE_STRING = "Charge Fouls"
PERSONAL_BLOCK_TYPE_STRING = "Personal Block Fouls"
PERSONAL_TAKE_TYPE_STRING = "Personal Take Fouls"
SHOOTING_BLOCK_TYPE_STRING = "Shooting Block Fouls"
CLEAR_PATH_FOUL_TYPE_STRING = "Clear Path Fouls"
DEFENSIVE_3_SECONDS_FOUL_TYPE_STRING = "Defensive 3 Seconds Violations"
FLAGRANT_1_FOUL_TYPE_STRING = "Flagrant 1 Fouls"
FLAGRANT_2_FOUL_TYPE_STRING = "Flagrant 2 Fouls"
DOUBLE_FOUL_TYPE_STRING = "Double Fouls"
INBOUND_FOUL_TYPE_STRING = "Inbound Fouls"

FOULS_DRAWN_TYPE_STRING = " Drawn"

OFF_DEADBALL_STRING = "OffDeadball"
OFF_LIVE_BALL_TURNOVER_STRING = "OffLiveBallTurnover"
OFF_TIMEOUT_STRING = "OffTimeout"
OFF_MADE_FT_STRING = f"Off{FREE_THROW_STRING}{MAKE_STRING}"
OFF_MISSED_FT_STRING = f"Off{FREE_THROW_STRING}{MISS_STRING}"

PENALTY_STRING = "Penalty"
FINAL_MINUTE_PENALTY_TAKE_FOUL_STRING = "FinalMinutePenaltyTakeFoul"
PENALTY_POSSESSIONS_EXCLUDING_TAKE_FOUL_FTS = "PenaltyOffPossExcludingTakeFouls"

PENALTY_FREE_THROW_STRING = "Penalty Free Throw Trips"
THREE_POINT_AND1_FREE_THROW_STRING = "3pt And 1 Free Throw Trips"
TWO_POINT_AND1_FREE_THROW_STRING = "2pt And 1 Free Throw Trips"
THREE_POINT_SHOOTING_FOUL_FREE_THROW_STRING = "3pt Shooting Foul Free Throw Trips"
TWO_POINT_SHOOTING_FOUL_FREE_THROW_STRING = "2pt Shooting Foul Free Throw Trips"
TECHNICAL_FREE_THROW_STRING = "Technical Free Throw Trips"

PLUS_MINUS_STRING = "PlusMinus"
TIME_STRING = "Time"

PLAYER_ID_STRING = "PlayerId"
TEAM_ID_STRING = "TeamId"
OPPONENT_TEAM_ID_STRING = "OpponentTeamId"
LINEUP_ID_STRING = "LineupId"
OPPONENT_LINEUP_ID_STRING = "OpponentLineupId"

TOTAL_2PT_SHOT_DISTANCE_STRING = "Total2ptShotDistance"
TOTAL_3PT_SHOT_DISTANCE_STRING = "Total3ptShotDistance"
TOTAL_2PT_SHOTS_WITH_DISTANCE = "Total2ptShotsWithDistance"
TOTAL_3PT_SHOTS_WITH_DISTANCE = "Total3ptShotsWithDistance"

LIVEBALL_TURNOVERS_STRING = "LiveBallTurnovers"
DEADBALL_TURNOVERS_STRING = "DeadBallTurnovers"
STEALS_STRING = "Steals"
TRAVELS_STRING = "Travels"
THREE_SECOND_VIOLATION_TURNOVER_STRING = "3SecondViolations"
SHOT_CLOCK_VIOLATION_TURNOVER_STRING = "ShotClockViolations"
STEP_OUT_OF_BOUNDS_TURNOVER_STRING = "StepOutOfBoundsTurnovers"
LOST_BALL_TURNOVER_STRING = "LostBallTurnovers"
LOST_BALL_OUT_OF_BOUNDS_TURNOVER_STRING = "LostBallOutOfBoundsTurnovers"
BAD_PASS_TURNOVER_STRING = "BadPassTurnovers"
BAD_PASS_OUT_OF_BOUNDS_TURNOVER_STRING = "BadPassOutOfBoundsTurnovers"
LOST_BALL_STEAL_STRING = "LostBallSteals"
BAD_PASS_STEAL_STRING = "BadPassSteals"


OFFENSIVE_GOALTENDING_STRING = "OffensiveGoaltends"
DEFENSIVE_GOALTENDING_STRING = "DefensiveGoaltends"


KEYS_OFF_BY_FACTOR_OF_5_WHEN_AGGREGATING_FOR_TEAM_AND_LINEUPS = [
    OFFENSIVE_POSSESSION_STRING,
    DEFENSIVE_POSSESSION_STRING,
    SECONDS_PLAYED_DEFENSE_STRING,
    SECONDS_PLAYED_OFFENSE_STRING,
    PLUS_MINUS_STRING,
    OPPONENT_POINTS,
    FREE_THROW_STRING + OFFENSIVE_ABBREVIATION_PREFIX + REBOUND_OPPORTUNITIES_STRING,
    AT_RIM_STRING + OFFENSIVE_ABBREVIATION_PREFIX + REBOUND_OPPORTUNITIES_STRING,
    SHORT_MID_RANGE_STRING
    + OFFENSIVE_ABBREVIATION_PREFIX
    + REBOUND_OPPORTUNITIES_STRING,
    LONG_MID_RANGE_STRING
    + OFFENSIVE_ABBREVIATION_PREFIX
    + REBOUND_OPPORTUNITIES_STRING,
    ARC_3_STRING + OFFENSIVE_ABBREVIATION_PREFIX + REBOUND_OPPORTUNITIES_STRING,
    CORNER_3_STRING + OFFENSIVE_ABBREVIATION_PREFIX + REBOUND_OPPORTUNITIES_STRING,
    AT_RIM_STRING
    + BLOCKED_STRING
    + OFFENSIVE_ABBREVIATION_PREFIX
    + REBOUND_OPPORTUNITIES_STRING,
    SHORT_MID_RANGE_STRING
    + BLOCKED_STRING
    + OFFENSIVE_ABBREVIATION_PREFIX
    + REBOUND_OPPORTUNITIES_STRING,
    LONG_MID_RANGE_STRING
    + BLOCKED_STRING
    + OFFENSIVE_ABBREVIATION_PREFIX
    + REBOUND_OPPORTUNITIES_STRING,
    ARC_3_STRING
    + BLOCKED_STRING
    + OFFENSIVE_ABBREVIATION_PREFIX
    + REBOUND_OPPORTUNITIES_STRING,
    CORNER_3_STRING
    + BLOCKED_STRING
    + OFFENSIVE_ABBREVIATION_PREFIX
    + REBOUND_OPPORTUNITIES_STRING,
    FREE_THROW_STRING + DEFENSIVE_ABBREVIATION_PREFIX + REBOUND_OPPORTUNITIES_STRING,
    AT_RIM_STRING + DEFENSIVE_ABBREVIATION_PREFIX + REBOUND_OPPORTUNITIES_STRING,
    SHORT_MID_RANGE_STRING
    + DEFENSIVE_ABBREVIATION_PREFIX
    + REBOUND_OPPORTUNITIES_STRING,
    LONG_MID_RANGE_STRING
    + DEFENSIVE_ABBREVIATION_PREFIX
    + REBOUND_OPPORTUNITIES_STRING,
    ARC_3_STRING + DEFENSIVE_ABBREVIATION_PREFIX + REBOUND_OPPORTUNITIES_STRING,
    CORNER_3_STRING + DEFENSIVE_ABBREVIATION_PREFIX + REBOUND_OPPORTUNITIES_STRING,
    AT_RIM_STRING
    + BLOCKED_STRING
    + DEFENSIVE_ABBREVIATION_PREFIX
    + REBOUND_OPPORTUNITIES_STRING,
    SHORT_MID_RANGE_STRING
    + BLOCKED_STRING
    + DEFENSIVE_ABBREVIATION_PREFIX
    + REBOUND_OPPORTUNITIES_STRING,
    LONG_MID_RANGE_STRING
    + BLOCKED_STRING
    + DEFENSIVE_ABBREVIATION_PREFIX
    + REBOUND_OPPORTUNITIES_STRING,
    ARC_3_STRING
    + BLOCKED_STRING
    + DEFENSIVE_ABBREVIATION_PREFIX
    + REBOUND_OPPORTUNITIES_STRING,
    CORNER_3_STRING
    + BLOCKED_STRING
    + DEFENSIVE_ABBREVIATION_PREFIX
    + REBOUND_OPPORTUNITIES_STRING,
    SECOND_CHANCE_STRING + OFFENSIVE_POSSESSION_STRING,
    SECOND_CHANCE_STRING + DEFENSIVE_POSSESSION_STRING,
    PENALTY_STRING + OFFENSIVE_POSSESSION_STRING,
    PENALTY_STRING + DEFENSIVE_POSSESSION_STRING,
    PENALTY_STRING + SECONDS_PLAYED_OFFENSE_STRING,
    PENALTY_STRING + SECONDS_PLAYED_DEFENSE_STRING,
    FINAL_MINUTE_PENALTY_TAKE_FOUL_STRING + OFFENSIVE_POSSESSION_STRING,
    FINAL_MINUTE_PENALTY_TAKE_FOUL_STRING + DEFENSIVE_POSSESSION_STRING,
    PENALTY_POSSESSIONS_EXCLUDING_TAKE_FOUL_FTS,
    PENALTY_STRING
    + FREE_THROW_STRING
    + OFFENSIVE_ABBREVIATION_PREFIX
    + REBOUND_OPPORTUNITIES_STRING,
    PENALTY_STRING
    + AT_RIM_STRING
    + OFFENSIVE_ABBREVIATION_PREFIX
    + REBOUND_OPPORTUNITIES_STRING,
    PENALTY_STRING
    + SHORT_MID_RANGE_STRING
    + OFFENSIVE_ABBREVIATION_PREFIX
    + REBOUND_OPPORTUNITIES_STRING,
    PENALTY_STRING
    + LONG_MID_RANGE_STRING
    + OFFENSIVE_ABBREVIATION_PREFIX
    + REBOUND_OPPORTUNITIES_STRING,
    PENALTY_STRING
    + ARC_3_STRING
    + OFFENSIVE_ABBREVIATION_PREFIX
    + REBOUND_OPPORTUNITIES_STRING,
    PENALTY_STRING
    + CORNER_3_STRING
    + OFFENSIVE_ABBREVIATION_PREFIX
    + REBOUND_OPPORTUNITIES_STRING,
    PENALTY_STRING
    + AT_RIM_STRING
    + BLOCKED_STRING
    + OFFENSIVE_ABBREVIATION_PREFIX
    + REBOUND_OPPORTUNITIES_STRING,
    PENALTY_STRING
    + SHORT_MID_RANGE_STRING
    + BLOCKED_STRING
    + OFFENSIVE_ABBREVIATION_PREFIX
    + REBOUND_OPPORTUNITIES_STRING,
    PENALTY_STRING
    + LONG_MID_RANGE_STRING
    + BLOCKED_STRING
    + OFFENSIVE_ABBREVIATION_PREFIX
    + REBOUND_OPPORTUNITIES_STRING,
    PENALTY_STRING
    + ARC_3_STRING
    + BLOCKED_STRING
    + OFFENSIVE_ABBREVIATION_PREFIX
    + REBOUND_OPPORTUNITIES_STRING,
    PENALTY_STRING
    + CORNER_3_STRING
    + BLOCKED_STRING
    + OFFENSIVE_ABBREVIATION_PREFIX
    + REBOUND_OPPORTUNITIES_STRING,
    PENALTY_STRING
    + FREE_THROW_STRING
    + DEFENSIVE_ABBREVIATION_PREFIX
    + REBOUND_OPPORTUNITIES_STRING,
    PENALTY_STRING
    + AT_RIM_STRING
    + DEFENSIVE_ABBREVIATION_PREFIX
    + REBOUND_OPPORTUNITIES_STRING,
    PENALTY_STRING
    + SHORT_MID_RANGE_STRING
    + DEFENSIVE_ABBREVIATION_PREFIX
    + REBOUND_OPPORTUNITIES_STRING,
    PENALTY_STRING
    + LONG_MID_RANGE_STRING
    + DEFENSIVE_ABBREVIATION_PREFIX
    + REBOUND_OPPORTUNITIES_STRING,
    PENALTY_STRING
    + ARC_3_STRING
    + DEFENSIVE_ABBREVIATION_PREFIX
    + REBOUND_OPPORTUNITIES_STRING,
    PENALTY_STRING
    + CORNER_3_STRING
    + DEFENSIVE_ABBREVIATION_PREFIX
    + REBOUND_OPPORTUNITIES_STRING,
    PENALTY_STRING
    + AT_RIM_STRING
    + BLOCKED_STRING
    + DEFENSIVE_ABBREVIATION_PREFIX
    + REBOUND_OPPORTUNITIES_STRING,
    PENALTY_STRING
    + SHORT_MID_RANGE_STRING
    + BLOCKED_STRING
    + DEFENSIVE_ABBREVIATION_PREFIX
    + REBOUND_OPPORTUNITIES_STRING,
    PENALTY_STRING
    + LONG_MID_RANGE_STRING
    + BLOCKED_STRING
    + DEFENSIVE_ABBREVIATION_PREFIX
    + REBOUND_OPPORTUNITIES_STRING,
    PENALTY_STRING
    + ARC_3_STRING
    + BLOCKED_STRING
    + DEFENSIVE_ABBREVIATION_PREFIX
    + REBOUND_OPPORTUNITIES_STRING,
    PENALTY_STRING
    + CORNER_3_STRING
    + BLOCKED_STRING
    + DEFENSIVE_ABBREVIATION_PREFIX
    + REBOUND_OPPORTUNITIES_STRING,
]



"""
``StatsNbaGameFinderLoader`` loads all games for a season and
creates :obj:`~pbpstats.resources.games.stats_nba_game_item.StatsNbaGameItem`
objects for each game

The following code will load data for the 2019-20 NBA Regular Season

.. code-block:: python

    from pbpstats.data_loader import StatsNbaGameFinderLoader

    game_finder_loader = StatsNbaGameFinderLoader("nba", "2019-20", "Regular Season", "web")
    print(game_finder_loader.items[0].data) # prints dict for first game
"""
import json
import os

from pbpstats import (
    NBA_STRING,
    G_LEAGUE_STRING,
    WNBA_STRING,
    NBA_GAME_ID_PREFIX,
    G_LEAGUE_GAME_ID_PREFIX,
    WNBA_GAME_ID_PREFIX,
)
from pbpstats.data_loader.abs_data_loader import check_file_directory
from pbpstats.data_loader.stats_nba.file_loader import StatsNbaFileLoader
from pbpstats.data_loader.stats_nba.web_loader import StatsNbaWebLoader
from pbpstats.resources.games.stats_nba_game_item import StatsNbaGameItem


class StatsNbaGameFinderLoader(StatsNbaFileLoader, StatsNbaWebLoader):
    """
    Loads stats.nba.com source data for season.
    Games are stored in items attribute
    as :obj:`~pbpstats.resources.games.stats_nba_game_item.StatsNbaGameItem` objects

    :param str league: Options are 'nba', 'wnba' or 'gleague'
    :param str season: Formatted as 2019-20 for NBA and G-League, 2019 of WNBA.
    :param str season_type: Options are 'Regular Season' or 'Playoffs' or 'Play In'
    :param str source: Where should data be loaded from. Options are 'web' or 'file'
    :param str file_directory: (optional if source is 'web')
        Directory in which data should be either stored (if source is web) or loaded from (if source is file).
        The specific file location will be `stats_<league>_<season>_<season_type>.json` in the `/schedule` subdirectory.
        If not provided response data will not be saved on disk.
    """

    data_provider = "stats_nba"
    resource = "Games"
    parent_object = "Season"

    def __init__(self, league, season, season_type, source, file_directory=None):
        self.league_string = league
        self.season_string = season
        self.season_type_string = season_type
        self.file_directory = file_directory
        self.source = source
        self._load_data()
        self._make_game_data_items()

    def _load_data(self):
        source_method = getattr(self, f"_from_{self.source}")
        source_method()

    @check_file_directory
    def _from_file(self):
        self.file_path = f'{self.file_directory}/schedule/stats_{self.league_string}_{self.season_string.replace("-", "_")}_{self.season_type_string.replace(" ", "_")}.json'
        self._load_data_from_file()

    def _from_web(self):
        league_url_part = (
            f"{G_LEAGUE_STRING}.{NBA_STRING}"
            if self.league_string == G_LEAGUE_STRING
            else self.league_string
        )
        self.base_url = f"https://stats.{league_url_part}.com/stats/leaguegamefinder"
        self.parameters = {
            "PlayerOrTeam": "T",
            "gtPTS": 1,
            "Season": self.season_string,
            "SeasonType": self.season_type_string,
            "LeagueID": self.league_id,
        }
        self._load_request_data()

    def _save_data_to_file(self):
        if self.file_directory is not None and os.path.isdir(self.file_directory):
            file_path = f'{self.file_directory}/schedule/stats_{self.league_string}_{self.season_string.replace("-", "_")}_{self.season_type_string.replace(" ", "_")}.json'
            with open(file_path, "w") as outfile:
                json.dump(self.source_data, outfile)

    @property
    def league_id(self):
        """
        Returns League Id for league.

        00 for nba, 10 for wnba, 20 for g-league
        """
        if self.league_string == NBA_STRING:
            return NBA_GAME_ID_PREFIX
        elif self.league_string == WNBA_STRING:
            return WNBA_GAME_ID_PREFIX
        elif self.league_string == G_LEAGUE_STRING:
            return G_LEAGUE_GAME_ID_PREFIX

    def _make_game_data_items(self):
        self.items = []
        sorted_games = sorted(self.data, key=lambda k: k["GAME_ID"])
        for team1, team2 in zip(sorted_games[0::2], sorted_games[1::2]):
            item = self._make_single_dict_for_game(team1, team2)
            self.items.append(StatsNbaGameItem(item))

    @staticmethod
    def _make_single_dict_for_game(team1, team2):
        """
        MATCHUP is vistor team @ home team for visitor
        MATCHUP is home team vs visitor team for home
        """
        item = {
            "GAME_ID": team1["GAME_ID"],
            "GAME_DATE_EST": team1["GAME_DATE"],
            "GAME_STATUS_TEXT": "Final",
            "HOME_TEAM_ID": team1["TEAM_ID"]
            if "@" not in team1["MATCHUP"]
            else team2["TEAM_ID"],
            "VISITOR_TEAM_ID": team1["TEAM_ID"]
            if "@" in team1["MATCHUP"]
            else team2["TEAM_ID"],
        }
        return item



"""
``StatsNbaShotsLoader`` loads shot data for a game and
creates :obj:`~pbpstats.resources.shots.stats_nba_shot.StatsNbaShot`
objects for all shots

The following code will load shot data for game id "0021900001" from
a file located in a subdirectory of the /data directory

.. code-block:: python

    from pbpstats.data_loader import StatsNbaShotsLoader

    shot_loader = StatsNbaShotsLoader("0021900001", "file", "/data")
    print(shot_loader.items[0].data) # prints dict with data for one shot from game
"""
import json
import os
import requests
from pathlib import Path

from pbpstats import HEADERS, REQUEST_TIMEOUT, NBA_STRING, G_LEAGUE_STRING
from pbpstats.data_loader.abs_data_loader import check_file_directory
from pbpstats.data_loader.stats_nba.web_loader import StatsNbaWebLoader
from pbpstats.data_loader.stats_nba.file_loader import StatsNbaFileLoader
from pbpstats.data_loader.stats_nba.summary_loader import StatsNbaSummaryLoader
from pbpstats.resources.shots.stats_nba_shot import StatsNbaShot


class StatsNbaShotsLoader(StatsNbaWebLoader, StatsNbaFileLoader):
    """
    Loads stats.nba.com source shot data for game.
    Shots are stored in items attribute
    as :obj:`~pbpstats.resources.shots.stats_nba_shot.StatsNbaShot` objects

    :param str game_id: NBA Stats Game Id
    :param str source: Where should data be loaded from. Options are 'web' or 'file'
    :param str file_directory: (optional if source is 'web')
        Directory in which data should be either stored (if source is web) or loaded from (if source is file).
        The specific file location will be `stats_home_shots_<game_id>.json`
        and `stats_away_shots_<game_id>.json` in the `/game_details` subdirectory.
        If not provided response data will not be saved on disk.
    """

    data_provider = "stats_nba"
    resource = "Shots"
    parent_object = "Game"

    def __init__(self, game_id, source, file_directory=None):
        self.game_id = game_id
        self.source = source
        self.file_directory = file_directory
        self._load_data()
        self._make_shot_items()

    def _load_data(self):
        source_method = getattr(self, f"_from_{self.source}")
        source_method()

    @check_file_directory
    def _from_file(self):
        self.home_file_path = (
            f"{self.file_directory }/game_details/stats_home_shots_{self.game_id}.json"
        )
        self.away_file_path = (
            f"{self.file_directory }/game_details/stats_away_shots_{self.game_id}.json"
        )

        home_data_file = Path(self.home_file_path)
        if not home_data_file.is_file():
            raise Exception(f"{self.home_file_path} does not exist")
        with open(self.home_file_path) as json_data:
            self.home_source_data = json.load(json_data)

        away_data_file = Path(self.away_file_path)
        if not away_data_file.is_file():
            raise Exception(f"{self.away_file_path} does not exist")
        with open(self.away_file_path) as json_data:
            self.away_source_data = json.load(json_data)

    def _from_web(self):
        summary = StatsNbaSummaryLoader(self.game_id, self.source, self.file_directory)
        home_team_id = summary.items[0].home_team_id
        away_team_id = summary.items[0].visitor_team_id
        league_url_part = (
            f"{G_LEAGUE_STRING}.{NBA_STRING}"
            if self.league == G_LEAGUE_STRING
            else self.league
        )
        self.base_url = f"https://stats.{league_url_part}.com/stats/shotchartdetail"
        base_paramaters = {
            "GameID": self.game_id,
            "Season": self.season,
            "SeasonType": self.season_type,
            "PlayerID": 0,
            "Outcome": "",
            "Location": "",
            "Month": 0,
            "SeasonSegment": "",
            "DateFrom": "",
            "DateTo": "",
            "OpponentTeamID": 0,
            "VsConference": "",
            "VsDivision": "",
            "Position": "",
            "RookieYear": "",
            "GameSegment": "",
            "Period": 0,
            "LastNGames": 0,
            "ContextMeasure": "FG_PCT",
            "PlayerPosition": "",
            "LeagueID": self.game_id[0:2],
        }
        self.home_parameters = base_paramaters.copy()
        self.home_parameters["TeamID"] = home_team_id
        self.away_parameters = base_paramaters.copy()
        self.away_parameters["TeamID"] = away_team_id

        response = requests.get(
            self.base_url,
            self.home_parameters,
            headers=HEADERS,
            timeout=REQUEST_TIMEOUT,
        )
        if response.status_code == 200:
            self.home_source_data = response.json()
        else:
            response.raise_for_status()

        response = requests.get(
            self.base_url,
            self.away_parameters,
            headers=HEADERS,
            timeout=REQUEST_TIMEOUT,
        )
        if response.status_code == 200:
            self.away_source_data = response.json()
        else:
            response.raise_for_status()
        self._save_data_to_file()

    def _save_data_to_file(self):
        if self.file_directory is not None and os.path.isdir(self.file_directory):
            home_file_path = f"{self.file_directory}/game_details/stats_home_shots_{self.game_id}.json"
            with open(home_file_path, "w") as outfile:
                json.dump(self.home_source_data, outfile)

            away_file_path = f"{self.file_directory}/game_details/stats_away_shots_{self.game_id}.json"
            with open(away_file_path, "w") as outfile:
                json.dump(self.away_source_data, outfile)

    def _make_shot_items(self):
        self.items = [StatsNbaShot(item) for item in self.data]

    def make_list_of_dicts(self, results_set_index=0):
        """
        Creates list of dicts from home and away source data

        :param int results_set_index: Index results are in. Default is 0
        :returns: list of dicts with shot data for all shots
        """
        headers = self.home_source_data["resultSets"][results_set_index]["headers"]
        home_rows = self.home_source_data["resultSets"][results_set_index]["rowSet"]
        home_deduped_rows = self.dedupe_events_row_set(home_rows)
        away_rows = self.away_source_data["resultSets"][results_set_index]["rowSet"]
        away_deduped_rows = self.dedupe_events_row_set(away_rows)
        return [dict(zip(headers, row)) for row in home_deduped_rows] + [
            dict(zip(headers, row)) for row in away_deduped_rows
        ]
